import JSBI from 'jsbi';

export type VoidCallback = () => void;
export type NumberCallback = (n: number) => void;
export type BufferCallback = (b: Buffer) => void;
export type StringCallback = (s: string) => void;

function throwInsufficientDataError(
  desiredRead: number,
  availableRead: number,
): never {
  throw new Error(
    `invalid read of size ${desiredRead}, only have ${availableRead} available`,
  );
}

/**
  A Buffer-like class that tracks position.

  As values are read, the position advances by the size of the read data.
  When reading, automatically verifies that enough data is available, and
  throws an error for insufficient buffered data.
 */
class ReadableTrackingBuffer {
  buffer: Buffer;

  position: number;

  constructor(buffer: Buffer) {
    this.buffer = Buffer.from(buffer);
    this.position = 0;
  }

  availableLength(): number {
    return this.buffer.length - this.position;
  }

  concat(buffer: Buffer) {
    if (this.position === this.buffer.length) {
      this.buffer = buffer;
    } else {
      this.buffer = Buffer.concat([
        this.buffer.slice(this.position),
        buffer,
      ]);
    }
    this.position = 0;
  }

  awaitData(length: number, callback: VoidCallback) {
    const availableLength = this.availableLength();
    if (length <= availableLength) {
      callback();
    } else {
      throwInsufficientDataError(length, availableLength);
    }
  }

  readInt8(callback: NumberCallback) {
    this.awaitData(1, () => {
      const data = this.buffer.readInt8(this.position);
      this.position += 1;
      callback(data);
    });
  }

  readUInt8(callback: NumberCallback) {
    this.awaitData(1, () => {
      const data = this.buffer.readUInt8(this.position);
      this.position += 1;
      callback(data);
    });
  }

  readInt16LE(callback: NumberCallback) {
    this.awaitData(2, () => {
      const data = this.buffer.readInt16LE(this.position);
      this.position += 2;
      callback(data);
    });
  }

  readInt16BE(callback: NumberCallback) {
    this.awaitData(2, () => {
      const data = this.buffer.readInt16BE(this.position);
      this.position += 2;
      callback(data);
    });
  }

  readUInt16LE(callback: NumberCallback) {
    this.awaitData(2, () => {
      const data = this.buffer.readUInt16LE(this.position);
      this.position += 2;
      callback(data);
    });
  }

  readUInt16BE(callback: NumberCallback) {
    this.awaitData(2, () => {
      const data = this.buffer.readUInt16BE(this.position);
      this.position += 2;
      callback(data);
    });
  }

  readInt32LE(callback: NumberCallback) {
    this.awaitData(4, () => {
      const data = this.buffer.readInt32LE(this.position);
      this.position += 4;
      callback(data);
    });
  }

  readInt32BE(callback: NumberCallback) {
    this.awaitData(4, () => {
      const data = this.buffer.readInt32BE(this.position);
      this.position += 4;
      callback(data);
    });
  }

  readUInt32LE(callback: NumberCallback) {
    this.awaitData(4, () => {
      const data = this.buffer.readUInt32LE(this.position);
      this.position += 4;
      callback(data);
    });
  }

  readUInt32BE(callback: NumberCallback) {
    this.awaitData(4, () => {
      const data = this.buffer.readUInt32BE(this.position);
      this.position += 4;
      callback(data);
    });
  }

  readBigInt64LE(callback: (data: JSBI) => void) {
    this.awaitData(8, () => {
      const result = JSBI.add(
        JSBI.leftShift(
          JSBI.BigInt(
            this.buffer[this.position + 4] +
            this.buffer[this.position + 5] * 2 ** 8 +
            this.buffer[this.position + 6] * 2 ** 16 +
            (this.buffer[this.position + 7] << 24) // Overflow
          ),
          JSBI.BigInt(32)
        ),
        JSBI.BigInt(
          this.buffer[this.position] +
          this.buffer[this.position + 1] * 2 ** 8 +
          this.buffer[this.position + 2] * 2 ** 16 +
          this.buffer[this.position + 3] * 2 ** 24
        )
      );

      this.position += 8;

      callback(result);
    });
  }

  readInt64LE(callback: NumberCallback) {
    this.awaitData(8, () => {
      const data = Math.pow(2, 32) * this.buffer.readInt32LE(this.position + 4) + ((this.buffer[this.position + 4] & 0x80) === 0x80 ? 1 : -1) * this.buffer.readUInt32LE(this.position);
      this.position += 8;
      callback(data);
    });
  }

  readInt64BE(callback: NumberCallback) {
    this.awaitData(8, () => {
      const data = Math.pow(2, 32) * this.buffer.readInt32BE(this.position) + ((this.buffer[this.position] & 0x80) === 0x80 ? 1 : -1) * this.buffer.readUInt32BE(this.position + 4);
      this.position += 8;
      callback(data);
    });
  }

  readBigUInt64LE(callback: (data: JSBI) => void) {
    this.awaitData(8, () => {
      const low = JSBI.BigInt(this.buffer.readUInt32LE(this.position));
      const high = JSBI.BigInt(this.buffer.readUInt32LE(this.position + 4));

      this.position += 8;

      callback(JSBI.add(low, JSBI.leftShift(high, JSBI.BigInt(32))));
    });
  }

  readUInt64LE(callback: NumberCallback) {
    this.awaitData(8, () => {
      const data = Math.pow(2, 32) * this.buffer.readUInt32LE(this.position + 4) + this.buffer.readUInt32LE(this.position);
      this.position += 8;
      callback(data);
    });
  }

  readUInt64BE(callback: NumberCallback) {
    this.awaitData(8, () => {
      const data = Math.pow(2, 32) * this.buffer.readUInt32BE(this.position) + this.buffer.readUInt32BE(this.position + 4);
      this.position += 8;
      callback(data);
    });
  }

  readFloatLE(callback: NumberCallback) {
    this.awaitData(4, () => {
      const data = this.buffer.readFloatLE(this.position);
      this.position += 4;
      callback(data);
    });
  }

  readFloatBE(callback: NumberCallback) {
    this.awaitData(4, () => {
      const data = this.buffer.readFloatBE(this.position);
      this.position += 4;
      callback(data);
    });
  }

  readDoubleLE(callback: NumberCallback) {
    this.awaitData(8, () => {
      const data = this.buffer.readDoubleLE(this.position);
      this.position += 8;
      callback(data);
    });
  }

  readDoubleBE(callback: NumberCallback) {
    this.awaitData(8, () => {
      const data = this.buffer.readDoubleBE(this.position);
      this.position += 8;
      callback(data);
    });
  }

  readUInt24LE(callback: NumberCallback) {
    this.awaitData(3, () => {
      const low = this.buffer.readUInt16LE(this.position);
      const high = this.buffer.readUInt8(this.position + 2);

      this.position += 3;

      callback(low | (high << 16));
    });
  }

  readUInt40LE(callback: NumberCallback) {
    this.awaitData(5, () => {
      const low = this.buffer.readUInt32LE(this.position);
      const high = this.buffer.readUInt8(this.position + 4);

      this.position += 5;

      callback((0x100000000 * high) + low);
    });
  }

  readUNumeric64LE(callback: NumberCallback) {
    this.awaitData(8, () => {
      const low = this.buffer.readUInt32LE(this.position);
      const high = this.buffer.readUInt32LE(this.position + 4);

      this.position += 8;

      callback((0x100000000 * high) + low);
    });
  }

  readUNumeric96LE(callback: NumberCallback) {
    this.awaitData(12, () => {
      const dword1 = this.buffer.readUInt32LE(this.position);
      const dword2 = this.buffer.readUInt32LE(this.position + 4);
      const dword3 = this.buffer.readUInt32LE(this.position + 8);

      this.position += 12;

      callback(dword1 + (0x100000000 * dword2) + (0x100000000 * 0x100000000 * dword3));
    });
  }

  readUNumeric128LE(callback: NumberCallback) {
    this.awaitData(16, () => {
      const dword1 = this.buffer.readUInt32LE(this.position);
      const dword2 = this.buffer.readUInt32LE(this.position + 4);
      const dword3 = this.buffer.readUInt32LE(this.position + 8);
      const dword4 = this.buffer.readUInt32LE(this.position + 12);

      this.position += 16;

      callback(dword1 + (0x100000000 * dword2) + (0x100000000 * 0x100000000 * dword3) + (0x100000000 * 0x100000000 * 0x100000000 * dword4));
    });
  }

  // Variable length data

  readBuffer(length: number, callback: BufferCallback) {
    this.awaitData(length, () => {
      const data = this.buffer.slice(this.position, this.position + length);
      this.position += length;
      callback(data);
    });
  }

  // Read a Unicode String (BVARCHAR)
  readBVarChar(callback: StringCallback) {
    this.readUInt8((length) => {
      this.readBuffer(length * 2, (data) => {
        callback(data.toString('ucs2'));
      });
    });
  }

  // Read a Unicode String (USVARCHAR)
  readUsVarChar(callback: StringCallback) {
    this.readUInt16LE((length) => {
      this.readBuffer(length * 2, (data) => {
        callback(data.toString('ucs2'));
      });
    });
  }

  // Read binary data (BVARBYTE)
  readBVarByte(callback: BufferCallback) {
    this.readUInt8((length) => {
      this.readBuffer(length, callback);
    });
  }

  // Read binary data (USVARBYTE)
  readUsVarByte(callback: BufferCallback) {
    this.readUInt16LE((length) => {
      this.readBuffer(length, callback);
    });
  }
}

export default ReadableTrackingBuffer;
module.exports = ReadableTrackingBuffer;
